pub mod state_dump;
mod timeouts;

use crate::{
    action::{Action, ActionWrapper},
    context::Context,
    instance::dispatch_action,
    state_dump::DumpOptions,
};
use std::sync::Arc;

pub fn create_state_dump_callback(
    context: Arc<Context>,
    options: DumpOptions,
) -> impl 'static + FnMut() + Sync + Send {
    move || {
        //log_debug!(context, "scheduled_jobs: tick");
        if context.state_dump_logging {
            state_dump::state_dump(context.clone(), options.clone());
        }
    }
}

pub fn create_timeout_callback(context: Arc<Context>) -> impl 'static + FnMut() + Sync + Send {
    move || {
        timeouts::check_network_processes_for_timeouts(context.clone());
    }
}

pub fn create_state_pruning_callback(
    context: Arc<Context>,
) -> impl 'static + FnMut() + Sync + Send {
    move || {
        dispatch_action(context.action_channel(), ActionWrapper::new(Action::Prune));
    }
}
